#include <math.h>
#include "Neighbor.h"
#include "Utils.h"


#define INC_BUFFER_DIST 1000
#define SIZE_BUFFER_TypeFloat 100
#define INC_BUFFER_ANC 10
#define NODE "NODE"
#define LEAF "LEAF"
#define MAXFLOAT 1000000000000.0

static TypeTreeNode *newNode(TypeNumber ident);
static void printNode(FILE *f, TypeTreeNode* node, char **name);
static void printNodeBis(FILE *f, TypeTreeNode* node, char **name, TypePosition depth);

void printTreeNexus(FILE *f, TypeNeighborTree tree) {
	TypeNumber i;
	printPreambleNexus(f, tree.name, tree.number);
	fprintf(f, "BEGIN Trees;\nproperties  partialtrees;\nTRANSLATE\n");
	for(i=0; i<tree.number; i++)
		fprintf(f, "\t'%s' '%s',\n", tree.name[i], tree.name[i]);
	fprintf(f, ";\nTREE 'Neighbor Joining' = ");
	printTreeNewick(f, tree);
	fprintf(f, ";\nEND;");
}

void printTreeNewick(FILE *f, TypeNeighborTree tree) {
	fprintf(f, "(");
	if(tree.root->child1 != NULL) {
		printNode(f, tree.root->child1, tree.name);
	}
	if(tree.root->child2 != NULL) {
		fprintf(f, ", ");
		printNode(f, tree.root->child2, tree.name);
	}
	if(tree.root->child3 != NULL) {
		fprintf(f, ", ");
		printNode(f, tree.root->child3, tree.name);
	}
	fprintf(f, ")");
}

/*print recursively the suffix node*/
void printNode(FILE *f, TypeTreeNode* node, char **name) {
	if(node == NULL)
		return;
	if(node->child1 != NULL) {
		fprintf(f, "(");
		printNode(f, node->child1, name);
		fprintf(f, ", ");
		printNode(f, node->child2, name);
		fprintf(f, "):");
		if(node->length>=0)
			fprintf(f, "%f", node->length);
		else
			fprintf(f, "0.0");
	} else {
		if(node->ident != -1)
			fprintf(f, "%s:%f", name[node->ident], node->length);
		else 
			fprintf(f, "problem");
	}
}

/*print recursively the suffix node*/
void printNodeBis(FILE *f, TypeTreeNode* node, char **name, TypePosition depth) {
	TypePosition d = depth;
	if(node == NULL)
		return;
	if(depth>=0) {
		/*Print the branches coming from higher nodes.*/
		while(d>0) {
			fprintf(f, "|");
			d--;
		}
		fprintf(f, "+");
		//fprintf(f, "[%f]-", node->length);
	}
	if(node->child1 != NULL) {
		fprintf(f, "\n");
		printNodeBis(f, node->child1, name, depth+1);
		printNodeBis(f, node->child2, name, depth+1);
		printNodeBis(f, node->child3, name, depth+1);
	} else {
		if(node->ident != -1)
			fprintf(f, "%s (%d)\n", name[node->ident], node->ident);
		else 
			fprintf(f, "problem\n");
	}
}

TypeNeighborTree computeTree(TypeDistance dist) {
	TypeNumber i, j, k, n;
	TypeFloat *R;
	TypeTreeNode **listnode;
	TypeNeighborTree tree;
	
	listnode = (TypeTreeNode**) monmalloc(dist.number*sizeof(TypeTreeNode*));
	R = (TypeFloat*) monmalloc(dist.number*sizeof(TypeFloat));

	for(i=0; i<dist.number; i++) {
		listnode[i] = newNode(i);
		R[i] = 0;
		for(j=0; j<i; j++) {
			R[i] += dist.table[(i*(i-1))/2+j];
		}
		for(j=i+1; j<dist.number; j++) {
			R[i] += dist.table[(j*(j-1))/2+i];
		}
	}
	for(n=dist.number; n>2; n--) {
		long i, j, k, imin, jmin;
		TypeFloat min, Dij;
		TypeTreeNode *newnode;

		min = MAXFLOAT;
		for(i=1; i<dist.number; i++) {
			if(listnode[i] != NULL) {
				for(j=0; j<i; j++) {
					if(listnode[j] != NULL) {
						TypeFloat Sij= (n-2)*dist.table[(i*(i-1))/2+j]-R[i]-R[j];
						if(Sij<min) {
							min = Sij;
							imin = i;
							jmin = j;
						}
					}
				}
			}
		}
		Dij = dist.table[(imin*(imin-1))/2+jmin];
		listnode[imin]->length = ((n-2)*Dij+R[imin]-R[jmin])/(2*(n-2));
		listnode[jmin]->length = Dij-listnode[imin]->length;
		if(listnode[imin]->length < 0.0) {
			listnode[imin]->length = 0.0;
			listnode[jmin]->length = (Dij>0)?Dij:0;
		} else if(listnode[jmin]->length < 0.0) {
			listnode[jmin]->length = 0.0;
			listnode[imin]->length = (Dij>0)?Dij:0;
		}
		newnode = newNode(-1);
		newnode->child1 = listnode[imin];
		newnode->child2 = listnode[jmin];
		listnode[imin] = newnode;
		listnode[jmin] = NULL;
/********************************/
/* update distances table and R */
/********************************/
		R[imin] = 0.0;
		for(k=0; k<jmin; k++) {
			if(listnode[k] != NULL) {
				TypeFloat Dik;
				R[k] -= dist.table[(imin*(imin-1))/2+k]+dist.table[(jmin*(jmin-1))/2+k];
				Dik = (dist.table[(imin*(imin-1))/2+k]+dist.table[(jmin*(jmin-1))/2+k]-Dij)/2;
				dist.table[(k*(k-1))/2+imin] = Dik;
				R[k] += Dik;
				R[imin] += Dik;
			}
		}
		for(k=jmin+1; k<imin; k++) {
			if(listnode[k] != NULL) {
				TypeFloat Dik;
				R[k] -= dist.table[(imin*(imin-1))/2+k]+dist.table[(k*(k-1))/2+jmin];
				Dik = (dist.table[(imin*(imin-1))/2+k]+dist.table[(k*(k-1))/2+jmin]-Dij)/2;
				dist.table[(imin*(imin-1))/2+k] = Dik;
				R[k] += Dik;
				R[imin] += Dik;
			}
		}
		for(k=imin+1; k<dist.number; k++) {
			if(listnode[k] != NULL) {
				TypeFloat Dik;
				R[k] -= dist.table[(k*(k-1))/2+imin]+dist.table[(k*(k-1))/2+jmin];
				Dik = (dist.table[(k*(k-1))/2+imin]+dist.table[(k*(k-1))/2+jmin]-Dij)/2;
				dist.table[(k*(k-1))/2+imin] = Dik;
				R[k] += Dik;
				R[imin] += Dik;
			}
		}
	}
	for(j=0; j<dist.number && listnode[j] == NULL; j++)
	;
	for(i=j+1; i<dist.number && listnode[i] == NULL; i++)
	;
printf("i %d j %d\n\n", i, j);
printNode(stdout, listnode[i], dist.name);
printf("\n\n");
printNode(stdout, listnode[j], dist.name);
printf("\n\n");
	listnode[j]->length = dist.table[(i*(i-1))/2+j];
	if(listnode[i]->child1 == NULL) {
		TypeTreeNode *newnode;
		newnode = newNode(-1);
		newnode->child1 = listnode[i];
		newnode->child2 = listnode[j];
		listnode[i] = newnode;
	} else {
		if(listnode[i]->child2 == NULL) {
			listnode[i]->child2 = listnode[j];
		} else {
			listnode[i]->child3 = listnode[j];
		}
	}
	monfree((void*) listnode);
	monfree((void*) R);
	tree.root = listnode[i];
	tree.name = dist.name;
	tree.number = dist.number;
	return tree;
}

TypeTreeNode *newNode(TypeNumber ident) {
	TypeTreeNode *node;
	node = (TypeTreeNode*) monmalloc(sizeof(TypeTreeNode));
	node->ident = ident;
	node->child1 = NULL;
	node->child2 = NULL;
	node->child3 = NULL;
	node->length = 0;
	return node;
}
