// 
// Interface between MS4 (Python) and NLD (C)
// 
// G. Grasseau (2010)
//

#include <Python.h>

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "Sequences.h"
#include "Distance.h"
#include "Decode.h"
#include "Neighbor.h"
#include "SequencesOld.h"
#include "Utils.h"
#include "gen.h"

#define ALPHABET_MAX  256
#define EMPTY         -1
#define DECIMALDIGITS 50   // Max digit number to decribe a decimal
#define CHARMAX       12   // The char max number of a GD-class label : 
                           // letter+6*digit+space


int ParseArguments( PyObject *args, TypeSetOfSequences *set, 
		    TypePosition *order_arg ) {


  PyObject *SeqsObj=NULL;

  //
  // Step 1 : Parse to Python sequence object + one integer
  //

  TypePosition order;
  if (!PyArg_ParseTuple( args, "Ol", &SeqsObj, &order)) 
		return 0;

  *order_arg = order;

  Py_ssize_t NbrOfSeq = PySequence_Size( SeqsObj );
  // fprintf( stderr, "Entering in parse: \n" );
  // fprintf( stderr, "  Nb Seq: %d\n", NbrOfSeq );


  //
  // Step 2 : build C structure out of Python object 
  //

  // 'code' translate character to alphabet ID 
  TypeSymbol code[ALPHABET_MAX];
  int i=0;
  for ( i=0; i < ALPHABET_MAX; i++)
    code[i] = EMPTY;
  
  // Initialise structure 'set'
  set->name  = malloc( sizeof(char *) * NbrOfSeq);
  set->table = NULL;
  set->title = '\0';
  set->number   = NbrOfSeq;
  set->cardinal = 0;
  set->size     = malloc(sizeof(TypePosition *) * NbrOfSeq);
  set->sequence = malloc(sizeof(TypeSymbol *)   * NbrOfSeq);
	
  PyObject *StringObj = NULL;
  char *string = NULL;

  // Position in sequence
  int pos = -1;

  // Sequence transformed into array of alphabet ID
  TypeSymbol *SeqIDs = NULL;
  int cardinal = 0;
  for ( i=0; i < NbrOfSeq; i++) {
    // fprintf( stderr, "Seq %d: \n", i );

    // Not used in NLD
    set->name[i] = NULL; 

    // Get sequence string
    StringObj   = PySequence_GetItem( SeqsObj, i );
    PyArg_Parse( StringObj, "s", &string );
    // fprintf( stderr, "Seq %s: \n", string );
    

    // Translate to NLD Identifier
    SeqIDs = malloc( sizeof(TypeSymbol) * ( strlen(string)+1 ) );
    pos = 0;
    char symbol = string[pos];
    while ( symbol != '\0' ) {
      symbol = toupper( symbol );
      if (code[ (int)symbol ] == EMPTY) {
	code[ (int)symbol ] = cardinal++;
      }
      SeqIDs[pos] = code[ (int) symbol ];
      pos++;
      symbol = string[pos];
    }
    set->size[i]     = pos;
    set->sequence[i] = SeqIDs;
  }
	
  // Set the conversion table Alphabet ID -> char  
  set->cardinal = cardinal;
  char *table = malloc( sizeof(char) * (cardinal+1) );
  for (i=0; i < ALPHABET_MAX; i++) {
    if (code[i] != EMPTY)
      table[ code[i] ] = (char) i;
  }
  
  table[cardinal]='\0';
  set->table=table;
  
  return 1;
}


// Build Python objects from ResSet (the decoded 
// sequences set) 
// The original sequence set (OrigSet) is required
//
PyObject * DeparseExact( TypeSetOfSequences ResSet,  
			  TypeSetOfSequences OrigSet
			 ) {
  int s;

  PyObject *ResultObj = PyList_New(0);

  // fprintf( stderr, "Deparse \n" );
  // fprintf( stderr, "  OrigSet %d: \n", OrigSet.number );
  // fprintf( stderr, "  ResSet  %d: \n", ResSet.number );
  
  // Numbering the same NLD-class
  TypeSymbol *SameClassID = (TypeSymbol*) 
    monmalloc( ResSet.cardinal * sizeof(TypeSymbol) );
  // ResSet.cardinal : alphabet size
  for( s=0; s < ResSet.cardinal; s++)
    SameClassID[s] = EMPTY;
  TypeSymbol *NbrOfSameClass = (TypeSymbol*) 
    monmalloc(OrigSet.cardinal*sizeof(TypeSymbol));
  for(s=0; s < OrigSet.cardinal; s++)
    NbrOfSameClass[s] = 0;
    
  int n = 0;
  char *buffer = NULL;
  char ClassName[CHARMAX];
  TypeSymbol ClassID;
  TypeSymbol OrigLetterID;
  for(n=0; n < OrigSet.number; n++) {
    TypePosition pos;
    buffer=calloc( CHARMAX * OrigSet.size[n], sizeof(char) );
    
    for(pos=0; pos < OrigSet.size[n]; pos++) {
      ClassID      = ResSet.sequence[n][pos];
      OrigLetterID = OrigSet.sequence[n][pos];
      if (SameClassID[ ClassID ] == EMPTY ) {
	SameClassID[ ClassID ] = NbrOfSameClass[ OrigLetterID ]++;
      }
      snprintf( ClassName, CHARMAX,"%c%d ", 
		OrigSet.table[ OrigLetterID ], 
		SameClassID[ ClassID ]);
      
      // fprintf( stderr, "  NLD-class %s: \n", ClassName );
      buffer = strncat( buffer, ClassName, CHARMAX);
    }
    PyObject *StringObj = PyString_FromString( buffer );
    free( buffer );
    PyList_Append( ResultObj, StringObj );		
  }
  
  // Free memory
  
  free( SameClassID );
  free( NbrOfSameClass );
  
  int index=0;
  for (index=0; index < OrigSet.number; index++) {
    free( ResSet.sequence[index] );
    if ( ResSet.name[index] != NULL )
      free( ResSet.name[index] );
    free( OrigSet.sequence[index] );
    if ( OrigSet.name[index] != NULL )
      free(OrigSet.name[index]);
  }
  
  //  OrigSet and ResSet use the same table, so only one is freed.	
  //  printf("ResSet.table: %p\tOrigSet.table: %p\n",ResSet.table,OrigSet.table);
  //  free(ResSet.table);

  
  free(ResSet.size);
  free(ResSet.sequence);

  free(OrigSet.table);
  free(OrigSet.size);
  free(OrigSet.sequence);
  free(OrigSet.name);
  
  return ResultObj;
}


// Build Python objects from ResSet (the decoded 
// sequences set) 
// The original sequence set (OrigSet) is required
//
PyObject * DeparseMismatch( Sequences ResSet ) {

  PyObject *ResultObj = PyList_New(0);

  // fprintf( stderr, "Deparse \n" );
  // fprintf( stderr, "  ResSet  %d: \n", ResSet.nombre );
  
  char *buffer = NULL;
  char ClassName[CHARMAX];
  
  long pos, n;
  for( n=0; n < ResSet.nombre; n++) {
    buffer=calloc( CHARMAX * ( ResSet.debut[n+1] - ResSet.debut[n]) , 
		   sizeof(char) );
    // fprintf( stderr, ">sequence %ld\n", n+1);
    for( pos=ResSet.debut[n]; pos < ResSet.debut[n+1]; pos++) {
      if(ResSet.sequence[pos] != EMPTY) {
	snprintf(ClassName, CHARMAX, "%ld ", ResSet.sequence[pos]);
	// fprintf( stderr, "Class %s \n", ClassName);
	buffer = strncat( buffer, ClassName, CHARMAX);
      }
    }
    PyObject *StringObj = PyString_FromString( buffer );
    free( buffer );
    PyList_Append( ResultObj, StringObj );		
  }

  return ResultObj;
}


PyObject *NLDMismatch( PyObject *self, PyObject *args ) {

  //
  // Step 1 : Parse arguments
  //
  TypeSetOfSequences SeqSet; 
  Sequences ConvertSet, *NMisSet, ResSet; 
  TypePosition order;

  ParseArguments( args, &SeqSet, &order);

  //
  // Step 2 : Compute NLD with mismatch
  //
  SeqSet.ambiguity.number = 0;
  ConvertSet = toOldSequences(SeqSet);
  freeSequences(SeqSet);

  NMisSet = Calcule_N_Mismatch(ConvertSet, order);

  ResSet  = Calcule_N_Antecedent_Mismatch(NMisSet, order);

  //
  // Step 3 : Retur result and deallocate memory
  //

  PyObject *ResultObj = DeparseMismatch( ResSet  ) ;

  return ResultObj;

}

PyObject *NLDExact( PyObject *self, PyObject *args ) {

  //
  // Step 1 : Parse arguments
  //

  // fprintf( stderr, "Entering: \n" );

  TypeSetOfSequences SeqSet;
  TypePosition order;

  // fprintf( stderr, " parseargs : %d \n", order );

  ParseArguments( args, &SeqSet, &order); 

  // fprintf( stderr, " seq size %d, order : %d \n", SeqSet.number, order );

  //
  // Step 2 : Compute exact NLD 
  //
  TypeSetOfSequences SeqSetOut = localDecode( SeqSet, order );

  //
  // Step 3 : Return result and deallocate memory
  //
  return ( DeparseExact( SeqSetOut, SeqSet ) );

}


//
//  Suffix Tree part
//


void storePySuffixNode(PyObject *PyObj, TypeSuffixNode* node, TypePosition depth) {
  TypePosition d = depth;
  TypeSymbol t;
  TypeSuffixNode *tmp;
	
  // Buffer depth + 2 * number_size + cst_characters
 
  char *buffer = malloc(depth + 2*DECIMALDIGITS + 5);
  buffer[0] = 0;
  char *str = malloc(DECIMALDIGITS);
  if(node == NULL) {

    return;
  }
  if(depth>=0) {
    // Print the branches coming from higher nodes.
    while(d>0) {
      buffer = strncat(buffer, "-", 1);
      d--;
    }
    buffer = strncat(buffer, "+", 1);
    snprintf(str, DECIMALDIGITS, "[%ld, ", node->start);
    buffer = strncat(buffer, str, DECIMALDIGITS);
    if(node->end != INFTY) {
      snprintf(str, DECIMALDIGITS, "%ld", node->end);
      buffer = strncat(buffer, str, DECIMALDIGITS);
    } else {
      snprintf(str, 6, "infty");
      buffer = strncat(buffer, str, 6);
    }
    snprintf(str, 2, "]\n");
    buffer = strncat(buffer, str, 2);
  }
  // Copy string
  PyObject *StringObj = PyString_FromString( buffer );
  free( str );
  free( buffer );

  PyList_Append( PyObj, StringObj );

  for(tmp=node->trans; tmp!=NULL; tmp = tmp->next) {
    storePySuffixNode(PyObj, tmp, depth+1);
  }
}

void storePySuffixTree(PyObject *PyObj, TypeSuffixTree *suffixTree) {
  storePySuffixNode(PyObj, suffixTree->root, 0);
}

PyObject *getSuffixTree( PyObject *self, PyObject *args ) {

  TypeSuffixTree *suffixTree;
  TypeOneSequence oneseq;
  TypeSetOfSequences SeqSet;
  TypePosition order;

  //
  // Step 1 : Parse to Python sequence object + one integer
  //

  ParseArguments( args, &SeqSet, &order); 

  // fprintf( stderr, "Entering in SuffixTree C: \n" );
  // fprintf( stderr, " order : %d \n", order );
  // fprintf( stderr, "  seq size: %d \n", SeqSet.number );

  oneseq = toOne(SeqSet);
  // printSequenceDebug( stderr,  oneseq, 80 );

  suffixTree = computeSuffixTree(oneseq);

  PyObject *PySuffixTree =  PyList_New(0);
  // ??
  // printSuffixTree( stderr, suffixTree );
  storePySuffixTree( PySuffixTree, suffixTree );
  freeSequences(SeqSet);
  freeSuffixTree( suffixTree );
  monfree((void*)oneseq.sequence);

  // fprintf( stderr, "Exiting in SuffixTree C: \n" );

  return ( PySuffixTree );
}


// Interface functions with python
static PyMethodDef MethodList[] = {

        // fonct. name, fonct. pointer, args. passing method, "comment"
	{ "NLDExact", NLDExact, METH_VARARGS, "NLD decoding"},
	{ "NLDMismatch", NLDMismatch, METH_VARARGS, "Mismatch NLD decoding"},
	{ "getSuffixTree", getSuffixTree, METH_VARARGS, "Get the Suffix Tree"},
	{NULL,NULL,0,NULL}
};

// Passes to Python MethodList 
PyMODINIT_FUNC
initMS4_lib(void)
{
  PyObject* m;
  m = Py_InitModule("MS4_lib", MethodList);
}


