#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <assert.h>
#include "Sequences.h"
#include "Utils.h"
#include "suffix-tree.h"

#define SPECIAL -1

#define LEFT 1
#define RIGHT 1 << 1
#define DONE 1 << 2
#define NEW 1 << 3


typedef struct TMPLOCALTREE {
	TypeLocalNode **buffer;
	TypePosition size, bufferSize, incSize;
} TypeTmpLocalTree;


static TypeSuffixNode *newSuffixNode(TypePosition start, TypePosition end);
static void freeNode(TypeSuffixNode *node);
static void printSuffixNode(FILE *f, TypeSuffixNode* node, TypePosition depth);
static void printLocalNode(FILE *f, TypeLocalNode* node, int *done, TypePosition depth);
static TypeSuffixNode *findTransition(TypeSuffixNode *node, TypeSymbol symbol, TypeSuffixTree *suffixTree);
static void addTransition(TypeSuffixNode *node, TypeSuffixNode *toadd, TypeSuffixTree *suffixTree);


static void canonize(TypeSuffixNode *node, TypePosition k, TypePosition p, TypeSuffixNode **sres, TypePosition *kres, TypeSuffixTree *suffixTree);
static void update(TypeSuffixNode *s, TypePosition k, TypePosition i, TypeSuffixNode **sres, TypePosition *kres, TypeSuffixTree *suffixTree);
static int test_and_split(TypeSuffixNode *s, TypePosition k, TypePosition p, TypeSymbol t, TypeSuffixTree *suffixTree, TypeSuffixNode **sres);

static TypeOneSequence computeSequence(TypeLocalTree *local);
static void writeSymbol(TypeLocalNode *node, TypeSymbol *sequence, TypeSymbol symb);
static void parse_and_check(TypeSuffixNode *suffixNode, TypeLocalTree *local, TypeTmpLocalTree *tmpLocal, TypePosition depth, TypePosition order);
static void parse_and_set(TypeSuffixNode *suffixNode, TypeLocalTree *local, TypeLocalNode *node, TypePosition depth);
static TypeLocalTree *initLocalTree(TypeSuffixTree *suffixTree, TypeOneSequence seq, TypePosition order);


//GG TypeLocalTree *iterateLocalTree(TypeTmpLocalTree *cur, TypeTmpLocalTree *new_);
void iterateLocalTree(TypeTmpLocalTree *cur, TypeTmpLocalTree *new_);
/*Compute local graph of the given order of positions of sequence seq
by computing the suffix tree of seq with Ukkonen's algorithm*/

/*free a node*/
void freeNode(TypeSuffixNode *node) {
	if(node != NULL) {
		TypeSuffixNode *tmp;
		tmp = node->trans;
		while(tmp != NULL) {
			TypeSuffixNode *retmp;
			retmp= tmp;
			tmp = tmp->next;
			freeNode(retmp);
		}
		monfree((void*)node);
	}
}
void freeSuffixTree(TypeSuffixTree *suffixTree) {
	freeNode(suffixTree->root);
	freeNode(suffixTree->bottom);
	monfree(suffixTree);
}

/*
void freeLocalTree(TypeLocalTree *localTree) {
	TypePosition pos;
	
	for(pos=0; pos<localTree->size; pos++) {
		if(localTree->leaf[pos] != NULL)
			freeLocalNode(getAncestor(localTree->leaf[pos]), localTree);
	}
	monfree((void*)localTree);
}

*/

TypeLocalNode* getAncestor(TypeLocalNode* node) {
	if(node->ancestor != NULL)
		return getAncestor(node->ancestor);
	else
		return node;
}/*test whether or not a state with canonical reference pair (s(k,p)) is the end-point (has a t transition i positions later)
unchanged from Ukkonen's algorithm*/
int test_and_split(TypeSuffixNode *s, TypePosition k, TypePosition p, TypeSymbol t, TypeSuffixTree *suffixTree, TypeSuffixNode **sres) {
	if(k<=p) {
		TypeSuffixNode **prev, *s2;
		for(prev = &(s->trans); *prev != NULL && suffixTree->sequence[(*prev)->start] < suffixTree->sequence[k]; prev = &((*prev)->next))
		;
		s2 = *prev;
		if(suffixTree->sequence[s2->start+p-k+1] == t) {
			*sres = s;
			return 1;
		} else {
			*prev = newSuffixNode(s2->start, s2->start+p-k);
			(*prev)->next = s2->next;
			s2->next = NULL;
			s2->start += p-k+1;
			(*prev)->trans = s2;
			*sres = *prev;
			return 0;
		}
	} else {
		*sres = s;
		return (findTransition(s, t, suffixTree) != NULL);
	}		
}
/*free a node*/
void freeLocalNode(TypeLocalNode *node, TypeLocalTree *localTree) {
	if(node != NULL) {
		TypeLocalNode *tmp;
		tmp = node->trans;
		if(tmp == NULL && localTree != NULL)
			localTree->leaf[node->position] = NULL;
		while(tmp != NULL) {
			TypeLocalNode *retmp;
			retmp= tmp;
			tmp = tmp->next;
			freeLocalNode(retmp, localTree);
		}
		monfree((void*)node);
	}
}
void addTransition(TypeSuffixNode *node, TypeSuffixNode *toadd, TypeSuffixTree *suffixTree) {
	TypeSuffixNode *cur;
	if(node->trans == NULL || suffixTree->sequence[node->trans->start]>suffixTree->sequence[toadd->start]) {
		toadd->next = node->trans;
		node->trans = toadd;
		return;
	}
	for(cur = node->trans; cur->next != NULL && suffixTree->sequence[cur->next->start]<suffixTree->sequence[toadd->start]; cur=cur->next)
	;
	toadd->next = cur->next;
	cur->next = toadd;
}	

TypeSuffixTree *computeSuffixTree(TypeOneSequence seq) {
	TypeSuffixNode *s;
	TypePosition i, k;
	TypeSuffixTree *suffixTree;
	
	suffixTree = (TypeSuffixTree*) monmalloc(sizeof(TypeSuffixTree));	
	suffixTree->bottom = newSuffixNode(0, -1);
	suffixTree->root = newSuffixNode(-1, -1);
	suffixTree->root->suffix = suffixTree->bottom;
	suffixTree->sequence = seq.sequence;
	s = suffixTree->root;
	k = 0;
	for(i=0; i<seq.size; i++) {
		update(s, k, i, &s, &k, suffixTree);
		canonize(s, k, i, &s, &k, suffixTree);
	}
	return suffixTree;
}
/*return the son of node with label starting by symbol if it exists, NULL otherwise*/
TypeSuffixNode *findTransition(TypeSuffixNode *node, TypeSymbol symbol, TypeSuffixTree *suffixTree) {
	TypeSuffixNode *tmp;
	if(node == suffixTree->bottom)
		return suffixTree->root;
	for(tmp=node->trans; tmp != NULL && suffixTree->sequence[tmp->start] < symbol; tmp=tmp->next)
	;
	if(tmp != NULL && suffixTree->sequence[tmp->start] == symbol)
		return tmp;
	return NULL;		
}

/*update the suffix suffixTree by adding the ith caracter of the sequence.
(s,(k,i-1)) is the canonical reference pair of the active point,
(sres, (kres,i)) return a reference pair of the next active point
suffixTree stores basic information relative to the sequence and the suffix suffixTree
depth is the depth of the active point (the length of the string spelled out from the root to this one)
order is the given order of decoding
link stores the local graph of this order
only changes from Ukkonen's algorithm are lines starting with /'star' 'star'/ 
concerning update of depth and local graph by adding links*/	
void update(TypeSuffixNode *s, TypePosition k, TypePosition i, TypeSuffixNode **sres, TypePosition *kres, TypeSuffixTree *suffixTree) {
	TypeSuffixNode *oldr, *r;
	int end_point;
	oldr = suffixTree->root;
	end_point = test_and_split(s, k, i-1, suffixTree->sequence[i], suffixTree, &r);
	while(!end_point) {
		addTransition(r, newSuffixNode(i, INFTY), suffixTree);
		if(oldr != suffixTree->root)
			oldr->suffix = r;
		oldr = r;
		canonize(s->suffix, k, i-1, &s, &k, suffixTree);
		end_point = test_and_split(s, k, i-1, suffixTree->sequence[i], suffixTree, &r);
	}
	if(oldr != suffixTree->root)
		oldr->suffix = s;
	*sres = s; *kres = k;
}

/*allocated and fill a new node*/
TypeSuffixNode *newSuffixNode(TypePosition start, TypePosition end) {
	TypeSuffixNode *node;
	node = (TypeSuffixNode*) monmalloc(sizeof(TypeSuffixNode));
	node->trans = NULL;
	node->next = NULL;
	node->suffix = NULL;
	node->start = start;
	node->end = end;
	return node;
}
/*return the canonical reference (sres,(kres,p)) of the state represented by the reference pair (s,(k,p))
unchanged from Ukkonen's algorithm*/
void canonize(TypeSuffixNode *s, TypePosition k, TypePosition p, TypeSuffixNode **sres, TypePosition *kres, TypeSuffixTree *suffixTree) {
	*sres = s;
	*kres = k;
	if(p>=k) {
		TypeSuffixNode *s2;
		s2 = findTransition(*sres, suffixTree->sequence[*kres], suffixTree);
		while(s2->end-s2->start <= p-*kres) {
			*kres += s2->end-s2->start+1;
			*sres = s2;
			if(*kres<=p)
				s2 = findTransition(*sres, suffixTree->sequence[*kres], suffixTree);
		}
	}
}

/*print the suffix tree*/
void printSuffixTree(FILE *f, TypeSuffixTree *suffixTree) {
	printSuffixNode(f, suffixTree->root, 0);
}

/*print recursively the suffix node*/
void printSuffixNode(FILE *f, TypeSuffixNode* node, TypePosition depth) {
	TypePosition d = depth;
	TypeSymbol t;
	TypeSuffixNode *tmp;
	
	if(node == NULL)
		return;
	if(depth>=0) {
		/*Print the branches coming from higher nodes.*/
		while(d>0) {
			fprintf(f, "-");
			d--;
		}
		fprintf(f, "+");
		fprintf(f, "[%ld, ", node->start);
		if(node->end != INFTY)
			fprintf(f, "%ld", node->end);
		else
			fprintf(f, "infty");
		fprintf(f, "]\n");
	}
	for(tmp=node->trans; tmp!=NULL; tmp = tmp->next) {
		printSuffixNode(f, tmp, depth+1);
	}
}

void getTree(TypeSetOfSequences set, TypePosition order) {
	FILE *fpSTree;
	TypeLink **link;
	TypeSuffixTree *suffixTree;
	TypeLocalTree *localTree;
	TypeOneSequence res;
	TypePosition pos;
	TypeOneSequence resOne, oneseq;

	oneseq = toOne(set);
	suffixTree = computeSuffixTree(oneseq);
	fpSTree = fopen("SuffixTree.out","w");
	assert(fpSTree);
	printSuffixTree(fpSTree,suffixTree);
	// localTree = initLocalTree(suffixTree, seq, order);
	freeSuffixTree(suffixTree);
	//	freeLocalTree(localTree);
}

