#! /usr/bin/env python

# -*- coding: iso-8859-1 -*-

##
##  Copyright (C) 2010 E. Corel, G. Didier, C. Devauchelle, G. Grasseau
##
##  This program is free software; you can redistribute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation; either version 2 of the License, or (at
##  your option) any later version.
## 
##  This program is distributed in the hope that it will be useful, but
##  WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
## 
##  You should have received a copy of the GNU General Public License
##  along with this program; if not, write to the Free Software
##  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
## 
##

# Importation de modules

import Tree, EquivClassesLayers
import re, sys, os, altgraph, itertools
from altgraph import Graph, GraphAlgo, Dot
from collections import deque
import string, time,pickle


## Concrete Class of hierachical Equivalent Classes.
#
#  A PartitionTree class deals with a hierachical tree which 
#  contains several ordered layers. Each layer corresponds to 
#  an N-order, and the last layer is compound of positions 
#  (seq_id, rel_pos) in the sequence (in fact the last layer is
#  the same that Equivalence Classes with an N-order equal 
#  to the size of the sequence).
#  So that, nodes of the PartitionTree are Equivalence Classes 
#  associated with an N-order. Each EC are included in a bigger one 
#  which have a smaller N-order
#   
class PartitionTree(Tree.Tree):


    ## Constructor.
    #  @param self the object pointer.
    #  @param ECLayers_ EquivClassLayers object
    def __init__(self, ECLayers_ = None):

        ## Root node of the Partition Tree
        self.root = None;

        ## dict() = [ nbr_of_seq, noduplication ]
        self.criterion = {}

        Tree.Tree.__init__(self)

        # ??? Tester le type
        if ECLayers_:
            self.initLeaves(ECLayers_.sites)
            self.addLayers(ECLayers_)
            self.root = self.getRoot()


    ## Initialise the leaves of the PartitionTree object with all sites of a SequenceSet
    def initLeaves(self, sites_ ):
        nbunch = []
        # Add leaves as couples (seq_id, rel_pos)
        for s in sites_.keys():     
            nbunch.extend([(s,i) for i in range(len(sites_[s]))])
        for u in nbunch:                
            self.add_node(u)            


    def getNmax(self,site):
        """Returns the Nmax of site"""
        node =self.inc_nbrs(site)[0]
        return int(self.node_data(node))

    def checkPTree(self, anchors=None):

        valid = True
        for n in anchors:
            pred = self.inc_nbrs( n )
            if( (self.criterion[n][1] != 1) and (len(pred) != 1) ):
                sys.stderr.write('Not valid anchors: '+ n + 
                                 'is a redondant EC or has several predecessors\n')    
                valid = False
            else:
                if( self.criterion[ pred[0] ][1] != 0 ):
                    sys.stderr.write('Not valid anchors: predecessor of '+ n + 
                                     'is a non-redondant EC\n')    
                    valid = False

        return ( valid )

    def writePTree(self, name_ = None, ext='', labels=None, tagnodes=None, detail=False):
        # Write the current PTree in ".dot" format
        # labels are the node names(string).
        # tagNodes are red colored
        # If "detail" is set, for each nodes the 2 criterion values are written: 
        # the occurence number of the curent EC and if the EC appears exactly 
        # once per sequence 

        dot = Dot.Dot(self)

        if not name_:
            name_ = 'ptree'

        if ( ( not labels ) or (len( labels ) != len( self.inner() )) ):
            for n in self.inner():
                if ( detail ):
                    str = "%s \\n %d %d" % (  n, self.criterion[n][0], int(self.criterion[n][1]) )
                else:
                    str = n
                if (n in set(tagnodes)):
                    style_ ='filled'
                    color_ = 'red'
                else:
                    style_ ='filled'
                    color_ = 'white'

                dot.node_style(n,fontsize='10',
                               label = str, style=style_, fillcolor = color_)
        #                       "\n %d %d" % (self.criterion[n]) )
        #            dot.node_style(n, style='filled', fillcolor = 'red')

        dot.save_dot(file_name=name_+ext)

    def saveTree(self,fileout=None,ext='.gvz'):
        """Produces a dot file with extension .gvz -- default 'tree.gvz'"""
        dot = Dot.Dot(self)
        for n in self.inner():
            dot.node_style(n,label = self.node_data(n))
        if not fileout:
            fileout = 'tree'
        dot.save_dot(file_name=fileout+ext)

# GG ???
#    def addLayers(self,dec,n,dico=None):

    # Adds layer of depth n to the current PTree as resulting from the decoding dec """
    # Must be orde
    def addLayers(self, ECLayers_ ):

        # From highest to lowest order
        for i in range(len( ECLayers_.order_list )):
            n = ECLayers_.order_list[i]
            newnodes = ECLayers_.dict_list[i]
            for nu in newnodes:
                nodes = set( [self.highestAncestor(v) 
                              for v in ECLayers_.dict_list[i][nu] ] )
                if len(nodes) > 1:
                    for a in nodes:
                        newNode = str(nu) + '_' + str( n )
                        self.add_node(newNode,n)
                        self.add_edge(newNode,a)
                        
        # Root level, 0-order
        n        = 0
        newnodes = [0]
        # GG ??? or ECLayers_.sites_
        nbunch = self.leaves()
        dict_0 = dict([(0,nbunch)])
        nodes  = set( [self.highestAncestor(v) 
                       for v in dict_0[0] ] )
        if len(nodes) > 1:
            for a in nodes:
                newNode = str(0) + '_' + str( n )
                self.add_node(newNode,n)
                self.add_edge(newNode,a)

    # Invalide : see next "prune" version 
    # GG - ???    def seqCounts(self,filt=1):
    # GG ??? 
    # Return a boolean dictionary of select nodes with the 
    # following criteria:
    # - the equivalence class must occur in all the sequences at least 
    #   occur_  times (??? attention occur n'est pas le smin)
    # - the equivalence class must appear once per sequence.
        """  Returns a boolean dictionary node -> False iff node has a repetition in the same sequence"""
    def prune_inv( self, occur_=1 ):
        root = self.root

        # Set root criterion
        self.criterion[ self.root ] = [0, 0]

        # To test the check
        # self.criterion[ self.root ] = [0, 1]

        flag = {}
        flag[root] = False
        for n in self.innerdfs():
            
            # Select leaves under (PT) node "n"
            leafList = self.childLeaves(n)

            # ??? list of sequences where n has a representative
            # Sequence index (ID) list of leaves (site list) under the node "n" (EC)
            seq_IDs = [pos[0] for pos in leafList] 

            # GG ??? number of representatives of seq in node 
            # Calculates the number of times where EC "n" appears in all sequences
            f = [seq_IDs.count(s) for s in set(seq_IDs)] 
            counts = reduce(lambda x,y:x+y,f)

            # ??? The node (EC) must appear exactly once per sequence.
            nonrep = (reduce(lambda x,y:x*y,f) == 1)

            flag[n] = (counts >= occur_ and nonrep)
            self.criterion[ n ] = [ counts, nonrep ]

        nodeList = self.getBorderNodes(lambda x:flag[x])

        return nodeList


    # GG - ???    def seqCounts(self,filt=1):
    # GG ??? 
    # Return a boolean dictionary of select nodes with the 
    # following criteria:
    # - the equivalence class must occur in all the sequences at least 
    #   occur_  times (??? attention occur n'est pas le smin)
    # - the equivalence class must appear once per sequence.
        """  Returns a boolean dictionary node -> False iff node has a repetition in the same sequence"""
    def prune( self, kappa = 1 ):
        root = self.root

        # Set root criterion
        self.criterion[ self.root ] = [0, 0]

        # To test the check
        # self.criterion[ self.root ] = [0, 1]

        flag = {}
        flag[root] = False
        for n in self.innerdfs():
            
            # Select leaves under (PT) node "n"
            leafList = self.childLeaves(n)

            # ??? list of sequences where n has a representative
            # Sequence index (ID) list of leaves (site list) under
            # the node "n" (EC)
            seq_IDs = [pos[0] for pos in leafList] 

            # GG ??? number of representatives of seq in node 
            # Calculates the number of times where EC "n" appears
            # in all sequences
            f = [seq_IDs.count(s) for s in set(seq_IDs)] 
            counts = reduce(lambda x,y:x+y,f)

            # "spread" is the number of sequences where the current
            # EC occures.
        
            spread = len( set( seq_IDs ) )
            
            flag[n] = ( ( float(counts) / spread ) <= kappa )
            self.criterion[ n ] = [ counts, spread ]

        # The largest EC such that [ (counts / spread) <= kappa]
        # are selected
        nodeList = self.getBorderNodes( lambda x:flag[x] )

        return nodeList
    
    # getMotif : to extract anchors occuring in "spread_filter" sequences
    # Not validated
    #
    # Return a boolean dictionary of select nodes with the 
    # following criteria:
    # - the equivalence class must occur in all the sequences at least 
    #   occur_  times (??? attention occur n'est pas le smin)
    # - the equivalence class must appear once per sequence.
        """  Returns a boolean dictionary node -> False iff node has a repetition in the same sequence"""
    def getMotif( self, spread_filter ):
        root = self.root

        # Set root criterion
        self.criterion[ self.root ] = [0, 0]

        # To test the check
        # self.criterion[ self.root ] = [0, 1]

        flag = {}
        flag[root] = False
        for n in self.innerdfs():
            
            # Select leaves under (PT) node "n"
            leafList = self.childLeaves(n)

            # ??? list of sequences where n has a representative
            # Sequence index (ID) list of leaves (site list) under
            # the node "n" (EC)
            seq_IDs = [pos[0] for pos in leafList] 

            # GG ??? number of representatives of seq in node 
            # Calculates the number of times where EC "n" appears
            # in all sequences
            f = [seq_IDs.count(s) for s in set(seq_IDs)] 
            counts = reduce(lambda x,y:x+y,f)

            # "spread" is the number of sequences where the current
            # EC occures.
        
            spread = len( set( seq_IDs ) )
            
            # flag[n] = ( ( float(counts) / spread ) <= kappa )
            flag[n] = ( spread >= spread_filter )
            self.criterion[ n ] = [ counts, spread ]

        # The largest EC such that [ (counts / spread) <= kappa]
        # are selected
        nodeList = self.getBorderNodes( lambda x:flag[x] )

        return nodeList


# ??? Pourquoi pas une methode de Partition Tree
# Given a list of nodes, compute the highest ancestor of node """
                   
def highestAncestor(tree,node):
    cv = node
    while True:
        try:
            # cv = tree.predecessors(cv)[0] #V1
            # Get parent node
            nb = tree.inc_nbrs(cv) #V0
            cv = nb[0] #V0
        except IndexError or TypeError:
                return cv

