#! /usr/bin/env python

# -*- coding: iso-8859-1 -*-

##
##  Copyright (C) 2010 E. Corel, G. Didier, C. Devauchelle, G. Grasseau
##
##  This program is free software; you can redistribute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation; either version 2 of the License, or (at
##  your option) any later version.
## 
##  This program is distributed in the hope that it will be useful, but
##  WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
## 
##  You should have received a copy of the GNU General Public License
##  along with this program; if not, write to the Free Software
##  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
## 
##

# Importation de modules

# import tG,mySyntax
import re, sys, os, altgraph,itertools
#import boost.graph as bgl
# import networkx as NX
from altgraph import Graph, GraphAlgo, Dot
from collections import deque
import string,time,pickle

##
# Tree
#

class Tree(Graph.Graph):

    def __init__(self,graph=None):
        """ Initialises the Tree structure -- empty by construction : NB the node_data must be the depth of the node (or any other increasing function) """
        if graph:
            Graph.Graph.__init__(self)
            for e in graph.edges:
                self.edges[e] = graph.edges[e][:]
            for n in graph.nodes:
                self.nodes[n] = graph.nodes[n][:]
            self.root = self.getRoot()
        else:
            Graph.Graph.__init__(self)
            self.root = None

    #========= Syntaxic methods
    def copyTree(self):
        """Looks like this copy style doesn't construct a real new independent copy ??"""
        t = Tree()
        for e in self.edges:
            try:
                head = self.edges[e][0][:]
            except TypeError:
                head = self.edges[e][0]
            try:
                tail = self.edges[e][1][:]
            except TypeError:
                tail = self.edges[e][1]
            t.edges[e] = (head,tail,self.edges[e][2])
        for n in self.nodes:
            t.nodes[n] = (self.nodes[n][0][:],self.nodes[n][1][:],self.nodes[n][2])
        try:
            t.root = self.root[:]
        except TypeError:
            t.root = self.root
        return t

    #========= Tree parsing methods

    def getRoot(self):
        """ Returns the root of tree self, or the list of roots, in case it is a forest """
        cv = self.nodes.keys().pop()
        while True:
            try:
                cv = self.inc_nbrs(cv)[0]
            except IndexError:
                return cv

    def getForestRoots(self):
        """ Returns the list of roots, in case it is a forest """
        roots = set([])
        for cv in self.leaves():
            #cv = self.nodes.keys().pop()
            while True:
                try:
                    cv = self.inc_nbrs(cv)[0]
                except IndexError:
                    roots.add(cv)
                    break
        if len(roots) == 1:
            roots = list(roots)[0]
        return roots

    def inner(self,start=1):
        """Generator of the inner node iterator of self -- start == 1 means start at root excluded -- obsolete"""
        if start:
            gen = (s for s in self.nodes if self.out_nbrs(s) and s != self.root)   #V1
        else:
            gen = (s for s in self.nodes if self.out_nbrs(s))   #V1
        while True:
            yield gen.next()

    def innerdfs(self,start='default'):
        """Generator of the inner node iterator of self on a DFS path --
        default start at root excluded;
        start == root means start at root included
        start == node means start at node included"""
        root = self.root
        if start == 'default':
            gen = (s for s in self.iterdfs(root) if self.out_nbrs(s) and s != root)   # starts from root but excludes it
        elif start == root:
            gen = (s for s in self.iterdfs(root) if self.out_nbrs(s))   # starts from root, included
        else:
            gen = (s for s in self.iterdfs(start) if self.out_nbrs(s))   # starts from start included
        while True:
            yield gen.next()

    def iterDFS(self,start='default'):
        """Generator of the inner node iterator of self on a DFS path --
        default start at root excluded;
        start == root means start at root included
        start == node means start at node included"""
        root = self.root
        if start == 'default':
            gen = (s for s in self.iterdfs(root) if s != root)   # starts from root but excludes it
        elif start == root:
            gen = (s for s in self.iterdfs(root))   # starts from root, included
        else:
            gen = (s for s in self.iterdfs(start))   # starts from start included
        while True:
            yield gen.next()

    def DFSprune(self,f,start=None):
        """Iterator on all nodes of tree pruned by condition f -- which is a (boolean) function defined on the nodes.
        Explicitly : starting from 0, outputs the nodes in DFS while condition is True.
        Note : if f is not monotonic, nothing tells that this returns all nodes satisfying f == True"""
        if not start:
            start = self.root
        # DFS starts here
        visited, stack = set([start]), deque([start])
        get_edges = self.out_edges
        while stack:
            curr_node = stack.pop()
            u = f(curr_node)
            if u != 0:
                if curr_node != start:
                    yield curr_node
                for edge in get_edges(curr_node):
                    tail = self.tail(edge)
                    if tail not in visited:
                       visited.add(tail)
                       if self.out_nbrs(tail):    # comment if you want to keep the leaves that are marked
                           stack.append(tail)
            else:
                # yield curr_node
                if curr_node not in visited:
                    visited.add(curr_node)

    def DFSpruneBorder(self,f,start=None):
        """Iterator on frontier of tree pruned by (increasing) condition -- outputs first node where condition f becomes True"""
        if not start:
            start = self.root
        # DFS starts here
        visited, stack = set([start]), deque([start])
        get_edges = self.out_edges
        while stack:
            curr_node = stack.pop()
            u = (f(curr_node) == 0)
            if u == 1:
                for edge in get_edges(curr_node):
                    tail = self.tail(edge)
                    if tail not in visited:
                       visited.add(tail)
                       if self.out_nbrs(tail):    # comment to include the marked leaves
                           stack.append(tail)
            else:
                yield curr_node
                if curr_node not in visited:
                    visited.add(curr_node)

    def DFSlast(self,f,start=None):
        """Iterator on frontier of tree pruned by (increasing) condition -- outputs last node where condition f is False: experimental"""
        if not start:
            start = self.root
        # DFS starts here
        visited, stack = set([start]), deque([start])
        get_edges = self.out_edges
        while stack:
            curr_node = stack.pop()
            u = (f(curr_node) == 0)
            print u
            if u:
                for edge in get_edges(curr_node):
                    tail = self.tail(edge)
                    if tail not in visited:
                       visited.add(tail)
                       if self.out_nbrs(tail):    # comment to include the marked leaves
                           stack.append(tail)
            else:
                yield stack[-1]      # yields last node before current
                if curr_node not in visited:
                    visited.add(curr_node)    

    def getInner(self,start=1):
        """List of inner nodes"""
        return [i for i in self.inner(start)]

    def leaves(self):
        """Generator of the leaf iterator of self"""
        gen = (s for s in self.nodes if not self.out_nbrs(s))      #V1
        while True:
            yield gen.next()

    def getLeaves(self):
        """List of leaves"""
        return [i for i in self.leaves()]

    def childLeaves(self, cv):
        """Extract the leaves under node cv"""
        if cv != self.root:
            d = []
            for s in self.iterdfs(cv):
                if not self.out_nbrs(s):     #V1
                    d.append(s)
        else:
            d = [s for s in self.leaves()]
        return d

    def iterAncestors(self,node):
        """ Iterator on the ancestors of a node """
        root = self.root
        father = lambda x:self.inc_nbrs(x)[0]
        if node != root:
            cv = node
            while cv != root:
                try:
                    nx = father(cv)
                    yield nx
                except TypeError:
                    print 'Erreur avec ',cv
                cv = nx
        else:
            yield root

    def ancestors(self,node):
        """Define the ancestors of node"""
        root = self.root
        if node != root:
            cv = node
            a = []
            while cv != root:
                try:
                    n = self.inc_nbrs(cv)[0]
                    a.extend(self.inc_nbrs(cv))
                except TypeError:
                    print 'Erreur avec ',cv
                cv = n
        else:
            a = [root]
        return a

    def nodeCounts(self):
        """Prints the number of nodes in self having a given depth
        and returns the dictionary i -> list of inner nodes having depth i"""
        c = {}
        for v in self.inner():
            i = int(self.node_data(v))
            try:
                c[i].append(v)
            except KeyError:
                c[i] = [v]
        return c

    #============= Decoding methods

    def highestAncestor(self,node):
        """For a forest, compute highest ancestor of node"""
        cv = node
        while True:
            try:
                nb = self.inc_nbrs(cv)
                cv = nb[0]
            except IndexError or TypeError:
                return cv

    def LCA(self,leafList):
        """Returns the least common ancestor of leafList in Tree self"""
        ga = reduce(lambda x,y:x.intersection(y),[set(self.ancestors(i)) for i in leafList])
        lga = list(ga)
        lga.sort(key=self.node_data)
        return lga[-1]

    def splitLeafList(self,leafList,k):
        """Returns the maximal subsets of leafList having a common ancestor of depth >= k -- The performance of this one is better"""
        child, anc = {},{}
        founders = set([])
        father = lambda x:self.inc_nbrs(x)[0]
        depth = self.node_data
        for leaf in leafList:
            cv = leaf
            n = father(cv)
            while depth(n) >= k:
                try:
                    child[n].append(leaf)
                    anc[leaf] = n
                except KeyError:
                    child[n] = [leaf]
                    anc[leaf] = n
                cv = n
                n = father(cv)
        founders = set(anc.values())
        return [child[key] for key in founders if len(child[key]) != 1]
        
    def testDFS(self,func):
        t0 = time.clock()
        for n in self.innerdfs():
            leafList = self.childLeaves(n)
            func(leafList)
        sys.stderr.write(func.__name__+' took '+str(time.clock()-t0)+'\n')    

    def minDepthNodes(self,n=1):
        """Returns iterator on nodes having depth immediately larger than n -- also to check with DFSpruneBorder"""
        a = []
        for i in self.iterdfs(0):
            try:
                if int(self.node_data(i))>=n and int(self.node_data(self.inc_nbrs(i)[0]))<n:
                    a.append(i)
            except TypeError:
                pass
        return iter(a)

    # GG ??? def treeFilter(self,function):
    def getNodes(self,function):
        """Returns the set of nodes filtered by function -- to compare with DFSprune"""
        nodeList = []
        for i in self.innerdfs():
            try:
                if function(i):
                    nodeList.append(i)
            except TypeError or KeyError:
                pass
        return nodeList

    # GG ??? def treeFilterBorder(self,function):    
    def getBorderNodes(self,function):
        """Returns the set of border nodes filtered by function -- to compare with DFSpruneBorder"""
        nodeList = []
        for i in self.innerdfs():
            try:
                if function(i) and not function(self.inc_nbrs(i)[0]):
                    nodeList.append(i)
            except TypeError or KeyError:
                pass
        return nodeList

    def saveTree(self,fileout=None,ext='.gvz'):
        """Produces a dot file with extension .gvz -- default 'tree.gvz'"""
        dot = Dot.Dot(self)
        for n in self.inner():
            dot.node_style(n,label = self.node_data(n))
        if not fileout:
            fileout = 'tree'
        dot.save_dot(file_name=fileout+ext)

    def viewTree(self,weights=True,nodeLabels=False):
        """ Visualises graph with edge weights -- version developed to by-pass bug in v. 2.16 of Graphviz """
        dot = Dot.Dot(self)
        if nodeLabels:
            for n in self.nodes:
                dot.node_style(n,label=self.node_data(n)[1])
        if weights:
            for e in self.edges:
                dot.edge_style(self.head(e),self.tail(e),label=self.edge_data(e))
        else:
            for e in self.edges:
                dot.edge_style(self.head(e),self.tail(e))
        dot.save_dot('tmp.dot')
        cmd = 'dot -Tpng tmp.dot > tmp.png; eog tmp.png'
        os.system(cmd)

# ex minDepthNodes with function lambda i:(int(self.node_data(i))>=n and int(self.node_data(self.inc_nbrs(i)[0]))<n)

#========== end of class definitions

def filtreMinDepth(node,tree,depth):
    return int(tree.node_data(node))>=depth and int(tree.node_data(tree.inc_nbrs(node)[0]))<depth

def readGraphviz(filename):
    """ Constructs a Graph.Graph() from a .dot file. """
    f = open(filename,'U')
    while True:    # skips the blank lines
        opening = f.readline()
        if opening:
            break
    if re.match('graph',opening):
        sep = '--'
    elif re.match('digraph',opening):
        sep = '->'
    else:
        sys.exit("Not a .dot file")
    #===================
    gr = Tree()
    elt = '(.*)\[(.*)\]'   # this matches the elements of graph
    edge = '(.*)'+sep+'(.*)\[(.*)\]'   # this matches specifically edges
    el = re.compile(elt) 
    ed = re.compile(edge) 
    for line in f.xreadlines():
        n = el.search(line)
        m = re.search(sep,line)
        if n:
            if not m:
                node = n.group(1).strip('\t').strip().strip('"')
                node = readNode(node)
                val = n.group(2).strip('label=').strip(',').strip('"')
                gr.add_node(node,val)
            else:
                e = ed.search(line)
                h,t = e.group(1).strip().strip('"'),e.group(2).strip().strip('"')
                h = readNode(h)
                t = readNode(t)
                if re.match('graph',opening):
                    gr.add_edge(h,t)
                    gr.add_edge(t,h)
                else:
                    gr.add_edge(h,t)
    f.close()
    return gr

def readNode(string):
    """Reads string to be put in graph as node, and tries to convert it to suitable types -- does nothing if fails"""
    try:
        node = int(string)
    except ValueError:
        try:
            node = mySyntax.tuplify(string)
        except (IndexError,ValueError):
            node = string
    return node


        
